/*
*	lr: list directories recursively
*	version: 1.3
*	date: 2013/05/30
*	auther: Kenar(Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*
*	compiling under plan9port
*		9c lr.c && 9l -o lr lr.o
*/

#include <u.h>
#include <libc.h>
#include <bio.h>

void
usage(void)
{
	fprint(2,"usage: lr [-Qlnwa] [-d depth] [-t time]  path ...\n"
	"lr lists directory recursively\n"
	"options:\n"
	"	-Q: unquoted format, instead TAB as separator (used only in combination with '-l')\n"
	"	-l: long format\n"
	"	-i: show date using iso-8601 style\n"
	"	-n: show date using unix epoch\n"
	"	-w: show warning for empty files\n"
	"	-d depth: search directory depth (positive integer)\n"
	"	-t time: show files modified since the time (seconds since the unix epoch)\n"
	"	-at time: show files accessed since the time (seconds since the unix epoch)\n"
	);
	exits("usage");
}

enum
{
	LFMT = 1,
	nFMT = 2,
	QFMT = 4,
	iFMT = 8,
};

char *mods[]={"---","--x","-w-","-wx","r--","r-x","rw-","rwx"};

int mode = 0;
Biobuf	out;
int wflag = 0; /* warning for empty files/dirs */
ulong stime = 0;
int aflag = 0; /* accessed time */

char*
modefmt(ulong m)
{
	static char buf[32];
	char *e, *s;
	s = buf;
	e = s + sizeof buf;
	s[0] = s[1] = '-';
	if(m & DMDIR)
		s[0] = 'd';
	else{
		if(m & DMAPPEND)
			s[0] = 'a';
		if(m & DMEXCL)
			s[1] = 'l';
	}
	s += 2;
	s = seprint(s,e, "%s", mods[(m & 0700) >> 6]);
	s = seprint(s,e, "%s", mods[(m & 070) >> 3]);
	s = seprint(s,e, "%s", mods[m & 07]);
	return buf;
}

char *
datefmt(ulong t)
{
	static char buf[48];
	Tm *tm;
	char sep = '/';

	if(mode & nFMT){
		snprint(buf,sizeof buf, "%lud", t);
		return buf;
	}
	tm = localtime(t);
	if(iFMT)
		sep = '-';
	snprint(buf,sizeof buf,"%04d%c%02d%c%02d %02d:%02d:%02d",
		tm->year + 1900,sep,tm->mon + 1,sep,tm->mday, tm->hour, tm->min, tm->sec);
	return buf;
}

/*
*	don't use d->name, the name is in path
*/
void
ls(char *path, Dir *d)
{
	if(mode & LFMT){
		if(mode & QFMT){
			Bprint(&out,"%s\t%s\t%s\t%10llud\t%s\t%s\n", modefmt(d->mode),
			d->uid, d->gid, d->length, datefmt(d->mtime), path);
		}
		else{
			Bprint(&out,"%s %-8s %-8s %10llud %s %q\n", modefmt(d->mode),
			d->uid, d->gid, d->length, datefmt(d->mtime), path);
		}
	}
	else{
		Bprint(&out,"%s", path);
		if(d->mode & DMDIR)
			Bprint(&out,"/");
		Bprint(&out,"\n");
	}
	if(wflag && d->length == 0 && (d->mode & DMDIR) == 0)
		fprint(2, "# empty file: %s\n", path);
}

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}


void
list(char *path, char *ep, int depth)
{
	Dir *db;
	int fd, i,n;
	ulong m;
	if(path[0] == '.' && path[1] == '/')
		path += 2;
	db = dirstat(path);
	if(db == nil){
		fprint(2, "# %s: non-existent or dead:%r\n", path);	
		return;
	}
	if(db->mtime > stime || (aflag && db->atime > stime))
		ls(path, db);
	m = db->mode;
	free(db);
	if((m & DMDIR) == 0)
		return;

	if(depth == 0)
		return;
	depth--;

	fd = open(path, OREAD);
	if(fd < 0){
		fprint(2, "# %s: unreadable\n", path);
		return;
	}
	n = dirreadall(fd, &db);
	close(fd);
	if(n < 0){
		fprint(2, "# %s: %r\n", path);
		return;
	}

	/* sort by name */
	qsort(db, n, sizeof(Dir), (int (*)(void*, void*))compar);
	//qsort(db, n, sizeof(Dir), (int (*)(const void*, const void*))compar);// for unix

	for(i = 0; i < n; i++){
		if(db[i].atime <= stime)
			continue;
		m = strlen(path);
		if(path[m-1] == '/')
			seprint(path + m - 1,ep,"/%s", db[i].name);
		else
			seprint(path + m,ep,"/%s", db[i].name);
		list(path, ep, depth);
		path[m] = 0;
	}
	free(db);
	if(wflag && n == 0)
		fprint(2, "# empty dir: %s\n", path);
}


void main(int argc, char *argv[])
{
	char path[4096];
	char *ep, *ap;
	int depth = -1;
	ARGBEGIN{
	case 'a':	aflag = 1; break;
	case 'd':
		ap = EARGF(usage());
		depth = atoi(ap);
		break;
	case 'l':	mode |= LFMT; break;
	case 'n':	mode |= nFMT; break;
	case 'Q':	mode |= QFMT; break;
	case 'i':	mode |= iFMT; break;
	case 't':
		ap = EARGF(usage());
		stime = (ulong)atol(ap);
		break;
	case 'w':	wflag = 1; break;
	default: usage();
	}ARGEND

	quotefmtinstall();
	Binit(&out, 1, OWRITE);
	ep = path + sizeof path;
	if(*argv == nil)
		usage();
	while(*argv){
		strecpy(path, ep, *argv);
		list(path, ep, depth);
		argv++;
	}
	exits(nil);
}
